<?php

    namespace FederationLib\Managers;

    use Doctrine\DBAL\ParameterType;
    use Exception;
    use FederationLib\Classes\Configuration;
    use FederationLib\Classes\Database;
    use FederationLib\Classes\Redis;
    use FederationLib\Classes\Security;
    use FederationLib\Classes\Utilities;
    use FederationLib\Classes\Validate;
    use FederationLib\Enums\DatabaseTables;
    use FederationLib\Enums\FilterOrder;
    use FederationLib\Enums\Filters\ListClientsFilter;
    use FederationLib\Exceptions\DatabaseException;
    use FederationLib\Exceptions\Standard\ClientNotFoundException;
    use FederationLib\Exceptions\Standard\InvalidClientDescriptionException;
    use FederationLib\Exceptions\Standard\InvalidClientNameException;
    use FederationLib\Exceptions\Standard\InvalidPermissionRoleException;
    use FederationLib\FederationLib;
    use FederationLib\Objects\Client;
    use LogLib\Log;
    use Symfony\Component\Uid\Uuid;
    use TamerLib\Enums\TamerMode;
    use TamerLib\tm;

    class ClientManager
    {
        /**
         * @var FederationLib
         */
        private FederationLib $federationLib;

        /**
         * ClientManager constructor.
         *
         * @param FederationLib $federationLib
         */
        public function __construct(FederationLib $federationLib)
        {
            $this->federationLib = $federationLib;
        }

        public function registerFunctions(): void
        {
            if(tm::getMode() !== TamerMode::WORKER)
            {
                return;
            }

            tm::addFunction('client_registerClient', [$this, 'registerClient']);
            tm::addFunction('client_getClient', [$this, 'getClient']);
            tm::addFunction('client_changeClientName', [$this, 'changeClientName']);
            tm::addFunction('client_changeClientDescription', [$this, 'changeClientDescription']);
            tm::addFunction('client_changeClientPermissionRole', [$this, 'changeClientPermissionRole']);
            tm::addFunction('client_updateClient', [$this, 'updateClient']);
            tm::addFunction('client_updateLastSeen', [$this, 'updateLastSeen']);
            tm::addFunction('client_listClients', [$this, 'listClients']);
            tm::addFunction('client_getTotalClients', [$this, 'getTotalClients']);
            tm::addFunction('client_getTotalPages', [$this, 'getTotalPages']);
            tm::addFunction('client_deleteClient', [$this, 'deleteClient']);
        }

        /**
         * Registers a client into the database, returns the UUID that was generated for the client.
         *
         * @param string|null $name
         * @param string|null $description
         * @return string
         * @throws DatabaseException
         * @throws InvalidClientDescriptionException
         * @throws InvalidClientNameException
         */
        public function registerClient(?string $name=null, ?string $description=null): string
        {
            $qb = Database::getConnection()->createQueryBuilder();
            $qb->insert(DatabaseTables::CLIENTS);
            $uuid = Uuid::v4()->toRfc4122();

            if($name === null)
            {
                $name = Utilities::generateName(4);
            }
            else
            {
                if(!Validate::clientName($name))
                {
                    throw new InvalidClientNameException(sprintf('Invalid client name: %s', $name));
                }
            }

            if($description !== null && strlen($description) > 128)
            {
                throw new InvalidClientDescriptionException(sprintf('Invalid client description: %s', $description));
            }

            $qb->setValue('uuid', $qb->createNamedParameter($uuid));
            $qb->setValue('name', $qb->createNamedParameter($name));
            $qb->setValue('description', $qb->createNamedParameter($description, (is_null($description) ? ParameterType::NULL : ParameterType::STRING)));
            $qb->setValue('secret_totp', $qb->createNamedParameter(Security::generateSecret()));
            $qb->setValue('flags', $qb->createNamedParameter(null, ParameterType::NULL));

            try
            {
                $qb->executeStatement();
            }
            catch(Exception $e)
            {
                throw new DatabaseException('Failed to register client: ' . $e->getMessage(), $e);
            }

            Log::info('net.nosial.federationlib', sprintf('Registered client with UUID %s', $uuid));
            return $uuid;
        }

        /**
         * Returns an existing client from the database.
         *
         * @param string|Client $client_uuid
         * @return Client
         * @throws ClientNotFoundException
         * @throws DatabaseException
         */
        public function getClient(string|Client $client_uuid): Client
        {
            if($client_uuid instanceof Client)
            {
                $client_uuid = $client_uuid->getUuid();
            }

            $qb = Database::getConnection()->createQueryBuilder();
            $qb->select('*');
            $qb->from(DatabaseTables::CLIENTS);
            $qb->where('uuid = :uuid');
            $qb->setParameter('uuid', $client_uuid);
            $qb->setMaxResults(1);

            try
            {
                $result = $qb->executeQuery();

                if($result->rowCount() === 0)
                {
                    throw new ClientNotFoundException($client_uuid);
                }

                $client = Client::fromArray($result->fetchAssociative());
            }
            catch(ClientNotFoundException $e)
            {
                throw $e;
            }
            catch(Exception $e)
            {
                throw new DatabaseException('Failed to get Client: ' . $e->getMessage(), $e);
            }

            return $client;
        }

        /**
         * Changes the name of a client.
         *
         * @param string|Client $client_uuid
         * @param string|null $name
         * @return void
         * @throws ClientNotFoundException
         * @throws DatabaseException
         * @throws InvalidClientNameException
         */
        public function changeClientName(string|Client $client_uuid, ?string $name=null): void
        {
            if($client_uuid instanceof Client)
            {
                $client_uuid = $client_uuid->getUuid();
            }

            if($name === null)
            {
                $name = Utilities::generateName(4);
            }
            else
            {
                if(!Validate::clientName($name))
                {
                    throw new InvalidClientNameException(sprintf('Invalid client name: %s', $name));
                }
            }

            $qb = Database::getConnection()->createQueryBuilder();
            $qb->update(DatabaseTables::CLIENTS);
            $qb->set('name', ':name');
            $qb->setParameter('name', $name);
            $qb->set('updated_timestamp', ':updated_timestamp');
            $qb->setParameter('updated_timestamp', time(), ParameterType::INTEGER);
            $qb->where('uuid = :uuid');
            $qb->setParameter('uuid', $client_uuid);
            $qb->setMaxResults(1);

            try
            {
                $affected_rows = $qb->executeStatement();
            }
            catch(Exception $e)
            {
                throw new DatabaseException('Failed to change client name: ' . $e->getMessage(), $e);
            }

            if($affected_rows === 0)
            {
                throw new ClientNotFoundException($client_uuid);
            }

            Log::verbose('net.nosial.federationlib', sprintf('Changed client name for client %s to %s', $client_uuid, $name));
        }

        /**
         * Changes the description of a client
         *
         * @param string|Client $client_uuid
         * @param string|null $description
         * @return void
         * @throws ClientNotFoundException
         * @throws DatabaseException
         * @throws InvalidClientDescriptionException
         */
        public function changeClientDescription(string|Client $client_uuid, ?string $description=null): void
        {
            if($client_uuid instanceof Client)
            {
                $client_uuid = $client_uuid->getUuid();
            }

            if($description !== null && strlen($description) > 128)
            {
                throw new InvalidClientDescriptionException(sprintf('Invalid client description: %s', $description));
            }

            $qb = Database::getConnection()->createQueryBuilder();
            $qb->update(DatabaseTables::CLIENTS);
            $qb->set('description', ':description');
            $qb->setParameter('description', $description, (is_null($description) ? ParameterType::NULL : ParameterType::STRING));
            $qb->set('updated_timestamp', ':updated_timestamp');
            $qb->setParameter('updated_timestamp', time(), ParameterType::INTEGER);
            $qb->where('uuid = :uuid');
            $qb->setParameter('uuid', $client_uuid);
            $qb->setMaxResults(1);

            try
            {
                $affected_rows = $qb->executeStatement();
            }
            catch(Exception $e)
            {
                throw new DatabaseException('Failed to change client description: ' . $e->getMessage(), $e);
            }

            if($affected_rows === 0)
            {
                throw new ClientNotFoundException($client_uuid);
            }

            Log::verbose('net.nosial.federationlib', sprintf('Changed client description for client %s to %s', $client_uuid, $description));
        }

        /**
         * Updates the permission role of a client.
         *
         * @param string|Client $client_uuid
         * @param int $permission_role
         * @return void
         * @throws ClientNotFoundException
         * @throws DatabaseException
         * @throws InvalidPermissionRoleException
         */
        public function changeClientPermissionRole(string|Client $client_uuid, int $permission_role): void
        {
            if($client_uuid instanceof  Client)
            {
                $client_uuid = $client_uuid->getUuid();
            }

            if(!Validate::permissionRole($permission_role))
            {
                throw new InvalidPermissionRoleException(sprintf('Invalid permission role: %s', $permission_role));
            }

            $time = time();

            $qb = Database::getConnection()->createQueryBuilder();
            $qb->update(DatabaseTables::CLIENTS);
            $qb->set('permission_role', ':permission_role');
            $qb->setParameter('permission_role', $permission_role, ParameterType::INTEGER);
            $qb->set('updated_timestamp', ':updated_timestamp');
            $qb->setParameter('updated_timestamp', $time, ParameterType::INTEGER);
            $qb->where('uuid = :uuid');
            $qb->setParameter('uuid', $client_uuid);
            $qb->setMaxResults(1);

            try
            {
                $affected_rows = $qb->executeStatement();
            }
            catch(Exception $e)
            {
                throw new DatabaseException('Failed to change client permission role: ' . $e->getMessage(), $e);
            }

            if($affected_rows === 0)
            {
                throw new ClientNotFoundException($client_uuid);
            }

            Log::verbose('net.nosial.federationlib', sprintf('Changed client permission role for client %s to %s', $client_uuid, $permission_role));
        }


        /**
         * Updates a client record in the database, if the client does not exist it will be created.
         * This function is cache aware, if the client is cached it will only update the changed values.
         *
         * @param Client $client
         * @return void
         * @throws DatabaseException
         */
        public function updateClient(Client $client): void
        {
            $cached_client = null;

            if(Configuration::isRedisCacheClientObjectsEnabled())
            {
                try
                {
                    if(Redis::getConnection()?->exists(sprintf('Client<%s>', $client->getUuid())))
                    {
                        $cached_client = Client::fromArray(Redis::getConnection()?->hGetAll(sprintf('Client<%s>', $client->getUuid())));
                    }
                }
                catch(Exception $e)
                {
                    Log::warning('net.nosial.federationlib', sprintf('Failed to get Client from redis: %s', $e->getMessage()));
                }
            }

            $qb = Database::getConnection()->createQueryBuilder();
            $qb->update(DatabaseTables::CLIENTS);
            $qb->set('updated_timestamp', ':updated_timestamp');
            $qb->setParameter('updated_timestamp', time());
            $qb->where('uuid = :uuid');
            $qb->setParameter('uuid', $client->getUuid());

            if($cached_client instanceof Client)
            {
                $data = array_diff($client->toArray(), $cached_client->toArray());
            }
            else
            {
                $data = $client->toArray();
            }

            foreach($data as $key => $value)
            {
                switch($key)
                {
                    case 'uuid':
                    case 'created_timestamp':
                    case 'updated_timestamp':
                    case 'seen_timestamp':
                        break;

                    case 'name':
                        if($value === null || strlen($value) === 0 || !preg_match('/^[a-zA-Z0-9_\-]+$/', $value ))
                        {
                            break;
                        }

                        $qb->set($key, ':' . $key);
                        $qb->setParameter($key, substr($value, 0, 64));
                        break;

                    case 'description':
                        if($value !== null)
                        {
                            $qb->set($key, ':' . $key);
                            $qb->setParameter($key, substr($value, 0, 255));
                        }
                        break;

                    case 'enabled':
                        $qb->set($key, ':' . $key);
                        $qb->setParameter($key, $value ? 1 : 0);
                        break;

                    default:
                        $qb->set($key, ':' . $key);
                        $qb->setParameter($key, $value);
                        break;
                }
            }

            try
            {
                $qb->executeStatement();
            }
            catch(Exception $e)
            {
                throw new DatabaseException('Failed to update client: ' . $e->getMessage(), $e);
            }

            if(Configuration::isRedisCacheClientObjectsEnabled())
            {
                // Update the differences in the cache
                if($cached_client instanceof Client)
                {
                    try
                    {
                        Redis::getConnection()?->hMSet((string)$client, array_diff($client->toArray(), $cached_client->toArray()));
                        Redis::getConnection()?->expire((string)$client, Configuration::getRedisCacheClientObjectsTTL());
                    }
                    catch(Exception $e)
                    {
                        Log::warning('net.nosial.federationlib', sprintf('Failed to cache client in redis: %s', $e->getMessage()));
                    }
                }
                else
                {
                    try
                    {
                        Redis::getConnection()?->hMSet((string)$client, $client->toArray());
                        Redis::getConnection()?->expire((string)$client, $client->getUuid(), Configuration::getRedisCacheClientObjectsTTL());
                    }
                    catch(Exception $e)
                    {
                        Log::warning('net.nosial.federationlib', sprintf('Failed to cache Client in redis: %s', $e->getMessage()));
                    }
                }
            }
        }

        /**
         * Updates a client's last seen timestamp.
         *
         * @param string|Client $uuid
         * @return void
         * @throws DatabaseException
         * @noinspection PhpUnused
         */
        public function updateLastSeen(string|Client $uuid): void
        {
            if($uuid instanceof Client)
            {
                $uuid = $uuid->getUuid();
            }

            $timestamp = time();

            $qb = Database::getConnection()->createQueryBuilder();
            $qb->update(DatabaseTables::CLIENTS);
            $qb->set('seen_timestamp', ':seen_timestamp');
            $qb->setParameter('seen_timestamp', $timestamp);
            $qb->where('uuid = :uuid');
            $qb->setParameter('uuid', $uuid);

            try
            {
                $qb->executeStatement();
            }
            catch(Exception $e)
            {
                throw new DatabaseException('Failed to update last seen timestamp: ' . $e->getMessage(), $e);
            }

            // Update the 'seen_timestamp' only in the hash table in redis
            if(Configuration::isRedisCacheClientObjectsEnabled())
            {
                try
                {
                    Redis::getConnection()?->hSet(sprintf('Client<%s>', $uuid), 'seen_timestamp', $timestamp);
                }
                catch(Exception $e)
                {
                    Log::warning('net.nosial.federationlib', sprintf('Failed to update last seen timestamp in redis: %s', $e->getMessage()));
                }
            }
        }

        /**
         * Returns an array of clients based on the filter, order, and page.
         *
         * @param int $page
         * @param string $filter
         * @param string $order
         * @param int $max_items
         * @return array
         * @throws DatabaseException
         */
        public function listClients(int $page, string $filter=ListClientsFilter::CREATED_TIMESTAMP, string $order=FilterOrder::ASCENDING, int $max_items=100): array
        {
            $qb = Database::getConnection()->createQueryBuilder();
            $qb->select(
                'uuid', 'enabled', 'name', 'description', 'secret_totp', 'query_permission', 'update_permission',
                'flags', 'created_timestamp', 'updated_timestamp', 'seen_timestamp'
            );
            $qb->from(DatabaseTables::CLIENTS);
            $qb->setFirstResult(($page - 1) * $max_items);
            $qb->setMaxResults($max_items);

            if($order !== FilterOrder::ASCENDING && $order !== FilterOrder::DESCENDING)
            {
                throw new DatabaseException('Invalid order: ' . $order);
            }

            switch($filter)
            {
                case ListClientsFilter::CREATED_TIMESTAMP:
                    $qb->orderBy('created_timestamp', strtoupper($order));
                    break;

                case ListClientsFilter::UPDATED_TIMESTAMP:
                    $qb->orderBy('updated_timestamp', strtoupper($order));
                    break;

                case ListClientsFilter::SEEN_TIMESTAMP:
                    $qb->orderBy('seen_timestamp', strtoupper($order));
                    break;

                case ListClientsFilter::ENABLED:
                    $qb->orderBy('enabled', strtoupper($order));
                    break;

                default:
                    throw new DatabaseException('Invalid filter: ' . $filter);
            }

            try
            {
                $result = $qb->executeQuery();
                $clients = [];

                while($row = $result->fetchAssociative())
                {
                    $clients[] = Client::fromArray($row);
                }
            }
            catch(Exception $e)
            {
                throw new DatabaseException('Failed to list clients: ' . $e->getMessage(), $e->getCode(), $e);
            }

            unset($client);
            return $clients;
        }

        /**
         * Returns the total number of clients.
         *
         * @return int
         * @throws DatabaseException
         */
        public function getTotalClients(): int
        {
            $qb = Database::getConnection()->createQueryBuilder();
            $qb->select('COUNT(uuid)');
            $qb->from(DatabaseTables::CLIENTS);

            try
            {
                $result = $qb->executeQuery();
                $row = $result->fetchAssociative();
            }
            catch(Exception $e)
            {
                throw new DatabaseException('Failed to get total clients: ' . $e->getMessage(), $e->getCode(), $e);
            }

            return (int)$row['COUNT(uuid)'];
        }

        /**
         * Returns teh total number of pages.
         *
         * @param int $max_items
         * @return int
         * @throws DatabaseException
         */
        public function getTotalPages(int $max_items=100): int
        {
            return (int)ceil($this->getTotalClients() / $max_items);
        }

        /**
         * Deletes an existing client from the database.
         *
         * @param string|Client $uuid
         * @return void
         * @throws DatabaseException
         */
        public function deleteClient(string|Client $uuid): void
        {
            if($uuid instanceof Client)
            {
                $uuid = $uuid->getUuid();
            }

            $qb = Database::getConnection()->createQueryBuilder();
            $qb->delete(DatabaseTables::CLIENTS);
            $qb->where('uuid = :uuid');
            $qb->setParameter('uuid', $uuid);

            try
            {
                $qb->executeStatement();
            }
            catch(Exception $e)
            {
                throw new DatabaseException('Failed to delete client: ' . $e->getMessage(), $e->getCode(), $e);
            }

            // Invalidate the cache
            if(Configuration::isRedisCacheClientObjectsEnabled())
            {
                try
                {
                    Redis::getConnection()?->del(sprintf('Client<%s>', $uuid));
                }
                catch(Exception $e)
                {
                    Log::warning('net.nosial.federationlib', sprintf('Failed to invalidate client cache in redis: %s', $e->getMessage()));
                }
            }
        }
    }